package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.connection.*;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import javafx.fxml.FXML;
import javafx.scene.layout.BorderPane;
import javafx.scene.web.WebEngine;
import javafx.scene.web.WebView;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.*;
import org.apache.http.entity.BasicHttpEntity;
import org.apache.http.impl.client.HttpClientBuilder;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.io.*;
import java.net.*;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;

public class GDPRDialog extends BorderPane {
    private static final Logger LOG = LoggerFactory.getLogger(GDPRDialog.class);

    @FXML private WebView reservationWebView;
    private WebEngine webEngine;
    private final GeniUserProvider geniUserProvider;
    private final SfaConnectionPool connectionPool;
    private final JFedTrustStore trustStore;

    public static final String EVENT_TYPE_CLICK = "click";

    @Inject
    public GDPRDialog(GeniUserProvider geniUserProvider,
                      SfaConnectionPool connectionPool,
                      JFedTrustStore trustStore) {
        this.geniUserProvider = geniUserProvider;
        this.connectionPool = connectionPool;
        this.trustStore = trustStore;
    }

    @FXML
    private void initialize() {
        webEngine = reservationWebView.getEngine();

        URLStreamFactoryCustomizer.useDedicatedProxyForWebkit(
                new MyHttpUrlStreamHandler(), "https");

        webEngine.load("https://www.wilab2.ilabt.iminds.be:12369/reservation/sfareservation.php3");
//        webEngine.load("https://jfed.ilabt.imec.be");
    }

    public HttpClientBuilder getHttpClientBuilder(URL url) {
        LOG.debug("getHttpClient url="+url);

        if (!geniUserProvider.isUserLoggedIn())
            return null;

        GeniUser user = geniUserProvider.getLoggedInGeniUser();
        BasicConnectionBuilderFactory connectionBuilderFactory = new BasicConnectionBuilderFactory();
        ConnectionBuilder connectionBuilder = connectionBuilderFactory.createConnectionBuilder();
        final JFedTrustStore conTrustStore = new JFedTrustStore(trustStore);
        Server server = null;
        //noinspection ConstantConditions
        if (server != null && server.hasCertificateChain()) {
            conTrustStore.addTrustedPemCertificateIfNotAdded(server.getCertificateChain());
            if (server.hasAllowedCertificateAlias()) {
                conTrustStore.addAllowedServerCertificateHostnameAlias(server.getAllowedCertificateAlias());
                LOG.debug("Using SSL conTrustStore with cert and with alias \""+server.getAllowedCertificateAlias()+"\" " +
                        "for server " + server.getName() + " (" + server.getId()+")");
            } else {
                LOG.debug("Using SSL conTrustStore with cert but without alias for server "+server.getName()+" ("+server.getId()+")");
            }
        }
//        else {
//            LOG.debug("Using SSL conTrustStore without cert or alias info for server "+server.getName()+" ("+server.getId()+")");
//        }
        connectionBuilder.useHttps(conTrustStore, null);
        connectionBuilder.useSslClientAuthentication(user.getClientCertificateChain(), user.getPrivateKey());
        connectionBuilder.setDebugInfo(new JFedConnection.DebugInfo(null, null));
        connectionBuilder.setUrl(url);
        HttpConnection con;
        try {
            con = connectionBuilder.buildHttpConnection();
        } catch (JFedException e) {
            LOG.error("Error building connection", e);
            return null;
        }

        HttpClientBuilder httpClientBuilder = ((BasicHttpConnection)con).getClientBuilder();
        return httpClientBuilder;
    }


    public class MyHttpUrlStreamHandler extends java.net.URLStreamHandler{
        @Override
        protected HttpURLConnection openConnection(URL u){
            LOG.debug("openConnection("+u.toExternalForm()+")");
            MyURLConnection q = new MyURLConnection(u);
            return q;
        }
    }

    private static class DebugOutputStream extends OutputStream {
        private final OutputStream delegate;

        private DebugOutputStream(OutputStream delegate) {
            this.delegate = delegate;
        }

        @Override
        public void write(int b) throws IOException {
            LOG.debug("write(b="+b+")");
            delegate.write(b);
        }

        @Override
        public void write(@NotNull byte[] b) throws IOException {
            LOG.debug("write(b[].length="+b.length+")");
            delegate.write(b);
        }

        @Override
        public void write(@NotNull byte[] b, int off, int len) throws IOException {
            LOG.debug("write(b[].length="+b.length+", off="+off+", len="+len+")");
            delegate.write(b, off, len);
        }

        @Override
        public void flush() throws IOException {
            LOG.debug("flush()");
            delegate.flush();
        }

        @Override
        public void close() throws IOException {
            LOG.debug("close()");
            delegate.close();
        }
    }

    private final Executor myHttpCLientExecutor = Executors.newSingleThreadExecutor();

    public class MyURLConnection extends HttpURLConnection
    {
        private final HttpClientBuilder httpClientBuilder;
        private HttpClient httpClient;
        private HttpResponse response;

        private final PipedOutputStream requestOutputStream;
        private final PipedInputStream requestInternalPipedInputStream;

        private final URI uri;

        private boolean responseError = false;

        protected MyURLConnection(URL url) {
            super(url);
            LOG.debug("MyURLConnection("+url+")");
            try {
                this.uri = url.toURI();
            } catch (URISyntaxException e) {
                throw new RuntimeException("Error creating URI from URL for \""+url.toExternalForm()+"\"", e);
            }
            this.httpClientBuilder = getHttpClientBuilder(url);
            this.requestOutputStream = new PipedOutputStream();
            this.requestInternalPipedInputStream = new PipedInputStream();
        }

        @Override
        public void disconnect() {
            LOG.debug("disconnect()");
            try {
                requestInternalPipedInputStream.close();
            } catch (IOException e) {
                //ignore any close exception
            }
            try {
                requestOutputStream.close();
            } catch (IOException e) {
                //ignore any close exception
            }
//            httpClient.?
//            response.?
        }

        @Override
        public boolean usingProxy() {
            LOG.debug("usingProxy()");
            return false;
        }

        @Override
        public void connect() throws IOException
        {
            LOG.debug("connect()");
            if (connected)
            {
                LOG.debug("already connected");
                return;
            }

            httpClientBuilder.setDefaultRequestConfig(RequestConfig.custom()
                    .setConnectTimeout(getConnectTimeout()) //setConnectTimeout = connection timeout in ms (default: 0 = infinite)
//                    .setConnectionRequestTimeout(getConnectTimeout()) //?
                    .setSocketTimeout(getReadTimeout()) //setSocketTimeout = read timeout in ms (default: 0 = infinite)
                    .build());

            this.httpClient = httpClientBuilder.build();


            HttpUriRequest httpCall;
            switch (method.toUpperCase()) {
                case "GET": { httpCall = new HttpGet(uri); break; }
                case "POST": { httpCall = new HttpPost(uri); break; }
                case "PUT": { httpCall = new HttpPut(uri); break; }
                case "DELETE": { httpCall = new HttpDelete(uri); break; }
                case "HEAD": { httpCall = new HttpHead(uri); break; }
                case "OPTIONS": { httpCall = new HttpOptions(uri); break; }
                case "TRACE": { httpCall = new HttpTrace(uri); break; }
                default: throw new ProtocolException("Unsupported HTTP method: " + method);
            }

            boolean onThread = false;
            if (httpCall instanceof HttpEntityEnclosingRequestBase) { //POST or PUT or ...
                BasicHttpEntity requestEntity = new BasicHttpEntity();
                if (fixedContentLength != -1)
                    requestEntity.setContentLength(fixedContentLength);
                if (fixedContentLengthLong != -1)
                    requestEntity.setContentLength(fixedContentLengthLong);
                requestEntity.setContent(requestInternalPipedInputStream);
                requestEntity.setChunked(chunkLength != -1); //we ignore the actual chunkLength, doesn't seem to be supported by httpClient
//                requestEntity.setContentType(???);
//                requestEntity.setContentEncoding(???);
                this.requestOutputStream.connect(requestInternalPipedInputStream);

                ((HttpEntityEnclosingRequestBase)httpCall).setEntity(requestEntity);

                onThread = true;
                LOG.debug("finished setup of requestEntity");
            }

            if (ifModifiedSince > 0) {
                Instant sinceDate = Instant.ofEpochMilli(ifModifiedSince);
                //If-Modified-Since expects RFC 822 (updated by RFC 1123)
                httpCall.setHeader("If-Modified-Since", java.time.format.DateTimeFormatter.RFC_1123_DATE_TIME.format(sinceDate));
            }

            Map<String, List<String>> requestHeaders = getRequestProperties();
            if (requestHeaders != null) {
                for (Map.Entry<String, List<String>> requestHeader : requestHeaders.entrySet()) {
                    httpCall.setHeader(requestHeader.getKey(), requestHeader.getValue().stream().collect(Collectors.joining(",")));
                }
            }

            connected = true;

            if (!onThread) {
                this.response = httpClient.execute(httpCall);
                LOG.debug("finished connect()");
            } else {
                Runnable exRun = () -> {
                    try {
                        this.response = httpClient.execute(httpCall);
                    } catch (IOException e) {
                        this.responseError = true;
                        throw new RuntimeException("httpClient.execute error", e);
                    }
                };
                myHttpCLientExecutor.execute(exRun);
                LOG.debug("finishing connect() in background thread");
            }
        }

        @Override
        public String getHeaderField(String name)
        {
            LOG.debug("getHeaderField("+name+")");
            waitForResponse();
            if (!response.containsHeader(name))
                return null;
            return response.getLastHeader(name).getValue();
        }

        @Override
        public String getContentType()
        {
            LOG.debug("getContentType()");
            waitForResponse();
            if (!response.containsHeader("Content-Type"))
                return null;
            return response.getLastHeader("Content-Type").getValue();
        }

        @Override
        public int getContentLength()
        {
            LOG.debug("getContentLength()");
            waitForResponse();
            return (int) response.getEntity().getContentLength();
        }

        @Override
        public long getContentLengthLong()
        {
            LOG.debug("getContentLengthLong()");
            waitForResponse();
            return response.getEntity().getContentLength();
        }

        @Override
        public boolean getDoInput()
        {
            LOG.debug("getDoInput()");
            return super.getDoInput();
        }

        @Override
        public InputStream getInputStream() throws IOException
        {
            connect();
            waitForResponse();
            LOG.debug("getInputStream()");
            return response.getEntity().getContent();
        }

        private void waitForResponse() {
            while (this.response == null && !responseError) {
                try {
                    LOG.debug("waitForResponse() is waiting");
                    Thread.sleep(100L);
                } catch (InterruptedException e) {
                    break; //break on interrupted
                }
            }
        }

        @Override
        public OutputStream getOutputStream() throws IOException
        {
            LOG.debug("getOutputStream()");
            return new DebugOutputStream(requestOutputStream);
        }

        @Override
        public java.security.Permission getPermission() throws IOException
        {
            LOG.debug("getPermission()");
            return super.getPermission();
        }

        @Override
        public String getHeaderFieldKey(int i) {
            LOG.debug("getHeaderFieldKey("+i+")");
            return super.getHeaderFieldKey(i);
        }

        @Override
        public void setFixedLengthStreamingMode(int i) {
            LOG.debug("setFixedLengthStreamingMode("+i+")");
            super.setFixedLengthStreamingMode(i);
        }

        @Override
        public void setFixedLengthStreamingMode(long l) {
            LOG.debug("setFixedLengthStreamingMode("+l+")");
            super.setFixedLengthStreamingMode(l);
        }

        @Override
        public void setChunkedStreamingMode(int i) {
            LOG.debug("setChunkedStreamingMode("+i+")");
            super.setChunkedStreamingMode(i);
        }

        @Override
        public String getHeaderField(int i) {
            LOG.debug("getHeaderField("+i+")");
            return super.getHeaderField(i);
        }

        @Override
        public void setInstanceFollowRedirects(boolean b) {
            LOG.debug("setInstanceFollowRedirects("+b+")");
            super.setInstanceFollowRedirects(b);
            if (!b)
                httpClientBuilder.disableRedirectHandling();
        }

        @Override
        public boolean getInstanceFollowRedirects() {
            LOG.debug("getInstanceFollowRedirects()");
            return super.getInstanceFollowRedirects();
        }

        @Override
        public void setRequestMethod(String s) throws ProtocolException {
            LOG.debug("setRequestMethod("+s+")");
            super.setRequestMethod(s);
        }

        @Override
        public String getRequestMethod() {
            LOG.debug("getRequestMethod()");
            return super.getRequestMethod();
        }

        @Override
        public int getResponseCode() throws IOException {
            LOG.debug("getResponseCode()");
            waitForResponse();
            return response.getStatusLine().getStatusCode();
//            return super.getResponseCode();
        }

        @Override
        public String getResponseMessage() throws IOException {
            LOG.debug("getResponseMessage()");
            waitForResponse();
            return response.getStatusLine().getReasonPhrase();
        }

        @Override
        public long getHeaderFieldDate(String s, long l) {
            LOG.debug("getHeaderFieldDate("+s+", "+l+")");
            //super calls getHeaderField so this will just work
            return super.getHeaderFieldDate(s, l);
        }

        @Override
        public InputStream getErrorStream() {
            LOG.debug("getErrorStream()");
            waitForResponse();
            return null; //TODO
            //probably something like this:
//            if (error && response.getEntity() != null)
//                return response.getEntity().getContent();
//            else
//                return null;
        }

        @Override
        public void setConnectTimeout(int i) {
            LOG.debug("setConnectTimeout("+i+")");
            super.setConnectTimeout(i);
        }

        @Override
        public int getConnectTimeout() {
            LOG.debug("getConnectTimeout() -> "+super.getConnectTimeout());
            return super.getConnectTimeout();
        }

        @Override
        public void setReadTimeout(int i) {
            LOG.debug("setReadTimeout("+i+")");
            super.setReadTimeout(i);
        }

        @Override
        public int getReadTimeout() {
            LOG.debug("getReadTimeout() -> "+super.getReadTimeout());
            return super.getReadTimeout();
        }

        @Override
        public URL getURL() {
            LOG.debug("getURL()");
            return super.getURL();
        }

        @Override
        public String getContentEncoding() {
            LOG.debug("getContentEncoding()");
            //just calls getHeaderField("content-encoding");
            return super.getContentEncoding();
        }

        @Override
        public long getExpiration() {
            LOG.debug("getExpiration()");
            //just calls getHeaderFieldDate("expires", 0);
            return super.getExpiration();
        }

        @Override
        public long getDate() {
            LOG.debug("getDate()");
            //just calls getHeaderFieldDate("date", 0);
            return super.getDate();
        }

        @Override
        public long getLastModified() {
            LOG.debug("getLastModified()");
            //just calls getHeaderFieldDate("last-modified", 0);
            return super.getLastModified();
        }

        @Override
        public Map<String, List<String>> getHeaderFields() {
            LOG.debug("getHeaderFields()");
            waitForResponse();
            Map<String, List<String>> res = new HashMap<>();
            for (Header header: response.getAllHeaders()) {
                res.put(header.getName(), Arrays.asList(header.getValue().split(",")));
            }
            return Collections.unmodifiableMap(res);
        }

        @Override
        public int getHeaderFieldInt(String s, int i) {
            LOG.debug("getHeaderFieldInt("+s+", "+i+")");
            //super uses getHeaderField(name); so ok to call
            return super.getHeaderFieldInt(s, i);
        }

        @Override
        public long getHeaderFieldLong(String s, long l) {
            LOG.debug("getHeaderFieldLong("+s+", "+l+")");
            //super uses getHeaderField(name); so ok to call
            return super.getHeaderFieldLong(s, l);
        }

        @Override
        public Object getContent() throws IOException {
            LOG.debug("getContent()");
            //super call should be ok to call as it uses getInputStream...
            return super.getContent();
        }

        @Override
        public Object getContent(Class[] classes) throws IOException {
            LOG.debug("getContent("+classes+")");
            //super call should be ok to call as it uses getInputStream...
            return super.getContent(classes);
        }

        @Override
        public void setDoInput(boolean b) {
            LOG.debug("setDoInput("+b+")");
            super.setDoInput(b);
        }

        @Override
        public void setDoOutput(boolean b) {
            LOG.debug("setDoOutput("+b+")");
            super.setDoOutput(b);
        }

        @Override
        public boolean getDoOutput() {
            LOG.debug("getDoOutput()");
            return super.getDoOutput();
        }

        @Override
        public void setAllowUserInteraction(boolean b) {
            LOG.debug("setAllowUserInteraction("+b+")");
            super.setAllowUserInteraction(b);
        }

        @Override
        public boolean getAllowUserInteraction() {
            LOG.debug("getAllowUserInteraction()");
            return super.getAllowUserInteraction();
        }

        @Override
        public void setUseCaches(boolean b) {
//            LOG.debug("setUseCaches("+b+")");
            super.setUseCaches(b);
        }

        @Override
        public boolean getUseCaches() {
//            LOG.debug("getUseCaches()");
            return super.getUseCaches();
        }

        @Override
        public void setIfModifiedSince(long l) {
            LOG.debug("setIfModifiedSince("+l+")");
            super.setIfModifiedSince(l);
        }

        @Override
        public long getIfModifiedSince() {
            LOG.debug("getIfModifiedSince()");
            return super.getIfModifiedSince();
        }

        @Override
        public boolean getDefaultUseCaches() {
            LOG.debug("getDefaultUseCaches()");
            return super.getDefaultUseCaches();
        }

        @Override
        public void setDefaultUseCaches(boolean b) {
            LOG.debug("setDefaultUseCaches("+b+")");
            super.setDefaultUseCaches(b);
        }

        @Override
        public void setRequestProperty(String s, String s1) {
            LOG.debug("setRequestProperty("+s+", "+s1+")");
            super.setRequestProperty(s, s1);
        }

        @Override
        public void addRequestProperty(String s, String s1) {
            LOG.debug("addRequestProperty("+s+", "+s1+")");
            super.addRequestProperty(s, s1);
        }

        @Override
        public String getRequestProperty(String s) {
            LOG.debug("getRequestProperty()");
            return super.getRequestProperty(s);
        }

        @Override
        public Map<String, List<String>> getRequestProperties() {
            LOG.debug("getRequestProperties() -> "+super.getRequestProperties());
            return super.getRequestProperties();
        }
    }
}
